//====----------------------- --- --   -      -
// MULTIUSER NETWORK MANAGEMENT
// TCP/IP and UDP support.
// 2000-2002, Vander Nunes
//======----------------------------- --- --   -      -


#include <windows.h>
#include "netclient.h"


DWORD CNetClient::m_dwRefCount = 0;


// ------------------------------------------------------------------------------------------------
//
// Client Window Procedure
//
// ------------------------------------------------------------------------------------------------
LRESULT CALLBACK NetClientWndProc(HWND hWnd, UINT uMessage, WPARAM wParam, LPARAM lParam)
{

  // get the pointer to the netserver class
  CNetClient *pClient = (CNetClient*)GetProp(hWnd, "netclient");
  if (!pClient)
    return DefWindowProc(hWnd,uMessage,wParam,lParam);

  // handle the message
  switch (uMessage)
  {

    case NET_SELECT:
    {
      pClient->Select(wParam, lParam);
      break;
    }

    case NET_RESOLVE:
    {
      if ((HANDLE)wParam == pClient->m_hResolver)
      {
        // this is our hostbyname result.
        int nRc = HIWORD(lParam);

        switch (nRc)
        {
          case 0:
          {
            // successful
            struct hostent *phe = (struct hostent *)pClient->m_szHostData;
            pClient->ContinueConnect(phe);
            break;
          }

          case WSAHOST_NOT_FOUND:
          {
            // authoritative: host not found
            break;
          }

          case WSATRY_AGAIN:
          {
            // non-authoritative host not found,
            // or server-fail.
            break;
          }

          default:
          {
            // there are more possible errors,
            // but we'll treat then as default.
            break;
          }

        }

      }
      break;
    }

  }

  // return default window procedure result
  return DefWindowProc(hWnd,uMessage,wParam,lParam);
}



// ------------------------------------------------------------------------------------------------
//
// Construction
//
// ------------------------------------------------------------------------------------------------
CNetClient::CNetClient(enSockType SockType, USR_CLNTCALLBACK(userProc), void *pUserData, int iBufferSize)
{
  m_bInitialized = FALSE;
  m_dwRefCount++;

  // try to initialize buffers
  if ( !Init(SockType, userProc, pUserData, iBufferSize) ) throw (0);

  // try to create hidden window
  if ( !(m_hWnd = NetMakeWindow("netclient",m_dwRefCount,64,32,WS_OVERLAPPEDWINDOW,NetClientWndProc)) ) throw (1);

  // put our address in the window properties
  if ( !SetProp(m_hWnd, "netclient", (HANDLE)this) ) throw (2);

  m_bInitialized = TRUE;
}



// ------------------------------------------------------------------------------------------------
//
// Initialization
//
// ------------------------------------------------------------------------------------------------
BOOL CNetClient::Init(enSockType SockType, USR_CLNTCALLBACK(userProc), void *pUserData, int iBufferSize)
{
  if (m_dwRefCount == 1)
  {
    // start winsock
    WSADATA wsaData;
    WORD wVersionRequested = MAKEWORD(1,1);
    WSAStartup(wVersionRequested, &wsaData);
  }

  // clear socket
  m_Sock = INVALID_SOCKET;

  // store connection information
  m_SockType = SockType;
  m_userProc = userProc;
  m_pUserData = pUserData;

  // initialize buffer
  m_iRecvBufferSize = iBufferSize;
  m_pRecvBuffer = new char[m_iRecvBufferSize];
  if (!m_pRecvBuffer) return FALSE;

  // point the packet buffer to the receive buffer
  m_Packet.pBuffer = (void*)m_pRecvBuffer;

  return TRUE;
}



// ------------------------------------------------------------------------------------------------
//
// Close winsock service
//
// ------------------------------------------------------------------------------------------------
CNetClient::~CNetClient()
{
  m_dwRefCount--;

  // disconnect
  Disconnect();

  // remove properties from our window
  RemoveProp(m_hWnd, "netclient");

  // destroy the window
  DestroyWindow(m_hWnd);

  if (!m_dwRefCount)
  {
    // close sockets service
    WSACleanup();
  }

  // free resources
  if (m_pRecvBuffer) delete m_pRecvBuffer;
}



// ------------------------------------------------------------------------------------------------
//
// Connect to a host
//
// ------------------------------------------------------------------------------------------------
BOOL CNetClient::Connect(char *szHost, int iPort)
{
  // refuses to connect if not initialized
  if (!m_bInitialized) return FALSE;

  // disconnect from any previous host
  Disconnect();

  // store port
  m_iPort = iPort;

  // initialize address
  memset((LPSTR)&m_MyAddr, 0, sizeof(m_MyAddr));
  m_MyAddr.sin_family = PF_INET;
  m_MyAddr.sin_port = htons(m_iPort);

  // resolve host
  memset(m_szHostData,0,sizeof(m_szHostData));
  m_hResolver = WSAAsyncGetHostByName(m_hWnd, NET_RESOLVE, szHost, m_szHostData, MAXGETHOSTSTRUCT);
  if (!m_hResolver)
    return FALSE;

  return TRUE;
}



// ------------------------------------------------------------------------------------------------
//
// Continue the connect process to a host
//
// ------------------------------------------------------------------------------------------------
BOOL CNetClient::ContinueConnect(struct hostent *phe)
{
  int nRc;

  // check if resolved correctly
  if (phe == NULL) return FALSE;

  // finish filling socket info
  if (phe->h_addr_list[0] != NULL)
    memcpy(&m_MyAddr.sin_addr, *(phe->h_addr_list), sizeof(*(phe->h_addr_list)));
  else
    return FALSE;

  // --------------------------------------------------------------------------
  // start connection
  // --------------------------------------------------------------------------

  // user wants to connect through UDP.
  if(m_Sock == INVALID_SOCKET)
  {

    if (m_SockType == stUDP)
    {
      // UDP
      m_Sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
    }
    else
      // TCP
      m_Sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);


    if(m_Sock != SOCKET_ERROR)
    {
      // set non-blocking bits
      unsigned long argp = 1;
      //ioctlsocket(m_Sock, FIONBIO, &argp);
    }
    else
    {
      // error while initializing UDP
      m_Sock = INVALID_SOCKET;
      return FALSE;
    }

  }

  // request an asynchronous connection
  nRc = WSAAsyncSelect(m_Sock,
      m_hWnd,                  // our window procedure will be the recipient
      NET_SELECT,              // message to send
      FD_CONNECT |             // tell when a connection is done
      FD_READ  |               // tell when new data's here
      FD_WRITE |               // tell when there's buffer space
      FD_CLOSE);               // tell if the other side closes

  // try connecting to host
  nRc = connect(m_Sock, (const struct sockaddr *)&m_MyAddr, sizeof(m_MyAddr));

  if (nRc == SOCKET_ERROR)
  {

    nRc = WSAGetLastError();
    switch (nRc)
    {
      default:
      {
        // really failed
        m_Sock = INVALID_SOCKET;
        return FALSE;
      }

      case WSAEWOULDBLOCK:
      {
        // ok, connect is queued...
        // just ignore it.
        break;
      }

    }

  }

  return TRUE;
}



// ------------------------------------------------------------------------------------------------
//
// Disconnect
//
// ------------------------------------------------------------------------------------------------
void CNetClient::Disconnect()
{
  // turn off UDP if it's running.
  if(m_Sock != INVALID_SOCKET)
  {
    closesocket(m_Sock);
    m_Sock = INVALID_SOCKET;
  }
}



// ------------------------------------------------------------------------------------------------
//
// Process connection events, that can be either TCP or UDP.
//
// ------------------------------------------------------------------------------------------------
BOOL CNetClient::Select(int wParam, int lParam)
{
  int nRc, i;
  unsigned long data = 0; 
  unsigned long four_byte;
  int namelen = sizeof(m_HostAddr), ilen = sizeof(unsigned long);

  // get the involved socket
  SOCKET sock = (SOCKET)wParam;

  switch(WSAGETSELECTEVENT(lParam))
  {

    case FD_CONNECT:
    {
      //
      // only TCP connections pass here
      //
      if (m_userProc) m_userProc(this, m_pUserData, NET_CONNECTED);
      break;
    }

    case FD_READ:
    {
      //
      // check if really there is need for reading
      //

      nRc = ioctlsocket(sock, FIONREAD, (u_long *)&data);
      if(nRc == SOCKET_ERROR)
        return FALSE;
      if(data <= 0 )
        return FALSE;

      // TCP doesn't fill in the from address on
      // recvfrom().  Find out if this is UDP or
      // tcp, and use an appropriate mechanism to
      // find out who the other end is.
      
      i = sizeof(int);
      getsockopt(sock, SOL_SOCKET, SO_TYPE, (char *)&nRc, (LPINT)&i);

      if(nRc == SOCK_DGRAM)
      {
        //
        // UDP datagram
        //
        nRc = recvfrom(sock, (LPSTR)&four_byte, (int)ilen, MSG_PEEK, (struct sockaddr *)&m_HostAddr, (LPINT)&namelen);
        if(nRc == SOCKET_ERROR)
        {
          nRc = WSAGetLastError();
          if(nRc != WSAEMSGSIZE)
            return FALSE;
        }
      }
      else
      {
        //
        // TCP packet
        //
        nRc = getpeername(sock, (struct sockaddr *)&m_HostAddr, (LPINT)&namelen);
      }
      if(nRc == SOCKET_ERROR)
        return FALSE;
      
      data = m_HostAddr.sin_addr.s_addr;

      // tell client there is data waiting
      if (m_userProc) m_userProc(this, m_pUserData, NET_NEWDATA);

      break;
    }

    case FD_CLOSE:
    {
      //
      // Only TCP pass here.
      //
      // the other side closed the connection.  close the socket,
      // to clean up internal resources.
      //

      // receive the disconnection message
      nRc = recvfrom(sock, (LPSTR)&four_byte, (int)ilen, MSG_PEEK, (struct sockaddr *)&m_HostAddr, (LPINT)&namelen);

      // if there is a last bit of data, let user read it.
      if (nRc > 0)
        if (m_userProc) m_userProc(this, m_pUserData, NET_NEWDATA);

      Disconnect();
      if (m_userProc) m_userProc(this, m_pUserData, NET_DISCONNECTION);

      break;
    }

  }   // end: switch(WSAGETSELECTEVENT(lParam))

  return TRUE;
}



// ------------------------------------------------------------------------------------------------
//
// Send a message to the host.
//
// ------------------------------------------------------------------------------------------------
void CNetClient::Send(void *Msg, int MsgLen)
{
  if (m_SockType == stTCP)
  {
    //
    // use TCP/IP send()
    //
    send(m_Sock, (const char *)Msg, MsgLen, 0);
  }
  else
  {
    //
    // use UDP sendto()
    //
    sendto(m_Sock, (const char *)Msg, MsgLen, 0, (struct sockaddr *)&m_HostAddr, sizeof(m_HostAddr));
  }
}



// ------------------------------------------------------------------------------------------------
//
// There is data waiting to be read.
// Receive that.
// Both TCP/IP and UDP pass here, after been 'selected'.
//
// ------------------------------------------------------------------------------------------------
NETPACKET* CNetClient::Read()
{
  int nRc, nTc=0;
  int ilen = sizeof(struct sockaddr_in);
  char *p = (char*)m_pRecvBuffer;

  do
  {

    if (m_SockType == stUDP)
    {
      nRc = recvfrom( m_Sock,
                      p,
                      m_iRecvBufferSize-nTc,
                      0,
                      (struct sockaddr *)&m_HostAddr,
                      (LPINT)&ilen
                    );
    }
    else
    {
      nRc = recv(m_Sock, p, m_iRecvBufferSize-nTc, 0);
    }

    if (nRc > 0)
    {
      nTc += nRc;
      p += nRc;
    }

  } while (nRc > 0);

  if (nTc > 0)
    m_Packet.iLength = nTc;
  else
    m_Packet.iLength = nRc;

  return &m_Packet;
}
